import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.applet.*;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Date;
import java.io.*;

/**
 * This is the Chatlet applet. It a simple user interface
 * with a common IRC client styled layout.
 *
 * @author Lars Samuelsson
 */
public class Chatlet extends Applet implements ConnectionHandler {
    // the font
    private Font typeface;
    // the chat layout
    private ChatPanel chat;
    // the login layout
    private LoginPanel login;
    // the nickname of the current user
    private String name;
    // the real name of the current user
    private String realname;
    // the address of the IRC server
    private String server;
    // the port of the IRC server
    private int port;
    // the URL for the servlet
    private String servletURL;
    // the identification code for this user
    private int code;
    // contains the complete graphical layout
    private CardLayout cards;
    // the connection
    private ServletConnection servlet;
    // an optional background image
    private Image background;
    // a flag indicating the connection status
    private boolean connected = false;
    // a message of the day that can be specified
    private String motd;
    // the font size
    private int fontSize;
    // a MediaTracker checking the progress of the background image
    private MediaTracker mediaTracker;
    // the background image id used by the MediaTracker
    private final int bgid = 0;
    // a flag indicating whether or not we want colors to be parsed
    private boolean parseColor;
    /**
     * Reverse delimiter.
     */
    public static final char reverse = (char) 0x16;
    /**
     * Underline delimiter.
     */
    public static final char underline = (char) 0x1f;
    /**
     * Color delimiter.
     */
    public static final char color = '\3';
    /**
     * Bold delimiter.
     */
    public static final char bold = '\2';
    /**
     * Used to convert a color string (fe84c2) to 
     * a Color object.
     * 
     * @param hexString The color given as a string 
     *                  of eight hexadecimal digits.
     */
    public Color hexColor(String hexString) {
	int r = Integer.decode("0x" + hexString.substring(0, 2)).intValue();
	int g = Integer.decode("0x" + hexString.substring(2, 4)).intValue();
	int b = Integer.decode("0x" + hexString.substring(4, 6)).intValue();
	return new Color(r, g, b);
    }
    /**
     * The initalization of the Chatlet is done here.
     * Parameters are read from the HTML code and then 
     * the layout is done using those values.
     */
    public void init() {
        System.out.println("This is Chatlet v1.0 - by lars@delicate.se");
	String param = getParameter("backgroundImage");
	if(param != null) {
	    mediaTracker = new MediaTracker(this);
	    background = getImage(getDocumentBase(), param);
	    mediaTracker.addImage(background, bgid);
	    if(background != null) {
		try {
		    mediaTracker.waitForID(bgid);
		}
		catch(InterruptedException e) {
		    // as good as done
		}
	    }
	}
	param = getParameter("parseColor");
	parseColor = !"false".equals(param);
	fontSize = 10;
	param = getParameter("fontSize");
	try {
	    fontSize = Integer.parseInt(param);
	}
	catch(Exception e) {
	    // disregard bad parameter
	}
	typeface = new Font("Dialog", Font.PLAIN, fontSize);
	motd = "";
	param = getParameter("motd");
	if(param != null)
	    motd = param;
	Color appletColor = Color.white;
	param = getParameter("appletColor");
	if(param != null)
	    appletColor = hexColor(param);
	if(background == null)
	    setBackground(appletColor);
	Color chatBackground = Color.white;
	param = getParameter("chatBackgroundColor");
	if(param != null)
	    chatBackground = hexColor(param);
	Color chatForeground = Color.black;
	param = getParameter("chatForegroundColor");
	if(param != null)
	    chatForeground = hexColor(param);
	Color inputBackground = Color.white;
	param = getParameter("inputBackgroundColor");
	if(param != null)
	    inputBackground = hexColor(param);
	Color inputForeground = Color.black;
	param = getParameter("inputForegroundColor");
	if(param != null)
	    inputForeground = hexColor(param);
        int historySize = 50;
        param = getParameter("historySize");
	try {
	    historySize = Integer.parseInt(param);
	}
        catch(Exception e) {
	    // disregard bad parameter
	}
	chat = new ChatPanel(this, 
			     typeface, 
			     chatBackground, 
			     chatForeground, 
			     inputBackground, 
			     inputForeground, 
			     historySize);
	Color labelForeground = Color.black;
	param = getParameter("labelForegroundColor");
	if(param != null) 
	    labelForeground = hexColor(param);
	String defaultServer = "";
	param = getParameter("defaultServer");
	if(param != null)
	    defaultServer = param;
	String defaultPort = "";
	param = getParameter("defaultPort");
	if(param != null)
	    defaultPort = param;
	param = getParameter("serverChange");
	boolean serverChange = !"false".equals(param);
	param = getParameter("portChange");
	boolean portChange = !"false".equals(param);
	login = new LoginPanel(this,
			       typeface,
			       labelForeground, 
			       inputBackground, 
			       inputForeground, 
			       defaultServer, 
			       serverChange,
			       defaultPort,
			       portChange);
	cards = new CardLayout();
	setLayout(cards);
	add("login", login);
	add("chat", chat);
	servletURL = getParameter("servletURL");
	servlet = new ServletConnection(this, servletURL);
    }
    /**
     * Reinitialize some things when started/restarted.
     */
    public void start() {
       chat.clear();
       cards.show(this, "login");
       login.repaint();
       showStatus("Chatlet Initialized");
    }
    /**
     * When the applet is stopped a disconnection is forced.
     */
    public void stop() {
	try {
	    servlet.writeObject(code + ":/quit");
	}
	catch(IOException e) {
	    // at least we tried to log out nicely
	}
	servlet.disconnect();
    }
    /**
     * Used to sound the system bell. 
     */
    public void beep() {
       try {
	  getToolkit().beep();
       }
       catch(Exception e) {
	  // alright, no beeping then
       }
    }
    /**
     * Used to append a String to the textfield in the chat.
     *
     * @param str The String to be appended.
     */
    public void append(String str) {
	if(parseColor) {
	   try {
	      str = parse(str);
	   }
	   catch(Exception e) {
	      System.out.println("Parsing error: " + e);
	   }
	}
	chat.append(str);
    }
    /**
     * Used when handshaking to get the connection data.
     * 
     * @return The connection data to use when handshaking.
     */
    public Object getConnectionData() {
	return server + " " + port + " " + name + " " + realname;
    }
    /**
     * The response given in a handshake procedure.
     * 
     * @param response The response from a handshake.
     */
    public void setConnectionResponse(Object response) {
	if(response instanceof String) { 
	   try {
	      code = Integer.parseInt(((String)response).trim());
	      connected = true;
	   }
	   catch(NumberFormatException e) {
	      showStatus("Bad code recieved");
	   }
	}
    }
    /**
     * Used to retrieve the current connection status.
     * 
     * @return true when a connection has been established
     *         and false otherwise.
     */
    public boolean isConnected() {
	return connected;
    }
    /**
     * Used to set the connection status.
     * 
     * @param connected true or false.
     */
    public void setConnected(boolean connected) {
	this.connected = connected;
        if(connected == false) {
	   showStatus("Connection lost");
	   cards.show(this, "login");
	}
    }
    /**
     * Used to handle data recieved by the ServletConnection.
     *
     * @param data The data to be handled.
     */
    public void handleData(Object data) {
       if(data != null) {
	  if(data.toString().startsWith("/")) {
	     execute(data.toString());
	  }
	  else {
	     append(data.toString());
	  }
       }
    }
    /**
     * Used to execute a command sent to the client from 
     * the servlet.
     * 
     * @param data The command data from the server.
     */
    public void execute(String data) {
       // ping request
       if(data.startsWith("/ping")) {
	  try {
	     servlet.writeObject(code + ":/pong");
	  }
	  catch(IOException e) {
	     // alright, no pong then
	  }
       }
    }
    /**
     * Used to parse color and bold controls in the 
     * incoming message.
     * 
     * @param str The unparsed String.
     * @return A String where color and bold controls 
     *         has been parsed.
     */
    public String parse(String str) throws NumberFormatException {
	int n = 0; // number of occurances
	int i = 0; // index of occurance
	String a, b;
	while((i = str.indexOf(bold, i)) > -1) {
	    a = str.substring(0, i++);
	    b = str.substring(i);
	    if(n++ % 2 == 0)
		str = a + "<bold>" + b;
	    else 
		str = a + "</bold>" + b;
	}
	i = 0;
	while((i = str.indexOf(reverse, i)) > -1) {
	    a = str.substring(0, i++);
	    b = str.substring(i);
	    if(n++ % 2 == 0)
		str = a + "<reverse>" + b;
	    else 
		str = a + "</reverse>" + b;
	}
	i = 0;
	while((i = str.indexOf(underline, i)) > -1) {
	    a = str.substring(0, i++);
	    b = str.substring(i);
	    if(n++ % 2 == 0)
		str = a + "<underline>" + b;
	    else 
		str = a + "</underline>" + b;
	}
        i = 0;
	while((i = str.indexOf(color, i)) > -1) {
	    a = str.substring(0, i++);
	    String fg = new String();
	    if(Character.isDigit(str.charAt(i))) {
		while(Character.isDigit(str.charAt(i)))
		    fg += str.charAt(i++);
		fg = " " + colorName(Integer.parseInt(fg));
	    }
	    String bg = new String();
	    if(str.charAt(i) == ',' && fg.length() > 0) {
		i++;
		while(Character.isDigit(str.charAt(i)))
		    bg += str.charAt(i++);
		bg = ", " + colorName(Integer.parseInt(bg));
	    }
	    b = str.substring(i);
	    if(n++ % 2 == 0)
		str = a + "<color" + fg + bg + ">" + b;
	    else
		str = a + "</color" + fg + bg + ">" + b;
	}
	return str;
    }
    /**
     * Gets the name of a color with a certain index.
     *
     * @param cidx The index of the color.
     * @return The name of the color as a String.
     */
    public String colorName(int cidx) {
	switch(cidx) {
	case  0: return "white";
	case  1: return "black";
	case  2: return "marine";
	case  3: return "forest";
	case  4: return "red";
	case  5: return "rust";
	case  6: return "purple";
	case  7: return "orange";
	case  8: return "yellow";
	case  9: return "green";
	case 10: return "slate";
	case 11: return "cyan";
	case 12: return "blue";
	case 13: return "pink";
	case 14: return "grey";
	case 15: return "steel";
	default: return "unknown";
	}
    }
    /**
     * A class used to control the maximum length of 
     * data inputted in a text component.
     * 
     * @author Lars Samuelsson
     */
    class InputLengthListener implements TextListener {
       int max;
       String text;
       public InputLengthListener(int max) {
	  this.max = max;
       }
       public void textValueChanged(TextEvent evt) {
	  TextComponent source = (TextComponent)evt.getSource();
	  text = source.getText();
	  if(text.length() > max) {
	     source.setText(text.substring(0, text.length() - 1));
	     source.setCaretPosition(text.length());
	     beep();
	  }
       }
    }
    /**
     * This panel is used in the interface when chatting.
     *
     * @author Lars Samuelsson
     */
    class ChatPanel extends Panel {
	// the area in which messages are shown
	private TextArea chatView;
	// the are where a user types a message or command
	private TextField input;
	// a reference to the chatlet
	private Chatlet chatlet;
	// the font used
	private Font typeface;
	// a history of recent input
	private History history;
	// current history position
	private int position = 0;
	/**
	 * This creates the ChatPanel.
	 *
	 * @param chatlet         A reference to the chatlet.
	 * @param typeface        The font to use.
	 * @param chatBackground  The background color of the message view.
	 * @param chatForeground  The foreground color of the message view.
	 * @param inputBackground The background of the input field.
	 * @param inputForeground The foreground of the input field.
	 * @param historySize     The size of the command history.
	 */
	public ChatPanel(Chatlet chatlet, 
			 Font typeface,
			 Color chatBackground, 
			 Color chatForeground, 
			 Color inputBackground, 
			 Color inputForeground, 
			 int historySize) {
	    this.chatlet = chatlet;
	    this.typeface = typeface;
	    history = new History(historySize);
	    GridBagLayout gbl = new GridBagLayout();
	    setLayout(gbl);
	    GridBagConstraints gbc = new GridBagConstraints();
	    gbl.setConstraints(this, gbc);
	    // layout for the chat view
	    gbc.weightx = 1;
	    gbc.weighty = 1;
	    gbc.gridx = 0;
	    gbc.gridy = 0;
	    gbc.fill = GridBagConstraints.BOTH;
	    gbc.insets = new Insets(5, 5, 5, 5);
	    chatView = new TextArea(chatlet.motd + "\n", 12, 60, 
				    TextArea.SCROLLBARS_VERTICAL_ONLY);
	    chatView.setFont(typeface);
	    chatView.setEditable(false);
	    chatView.setBackground(chatBackground);
	    chatView.setForeground(chatForeground);
	    add(chatView, gbc);
	    // layout for the input field
	    gbc.gridx = 0; 
	    gbc.gridy = 1;
	    gbc.weightx = 1;
	    gbc.weighty = 0;
	    gbc.insets = new Insets(0, 5, 5, 5);
	    gbc.fill = GridBagConstraints.HORIZONTAL;
	    input = new TextField(30);
	    input.setFont(typeface);
	    input.addActionListener(new InputListener());
	    input.addKeyListener(new HistoryListener());
	    input.addTextListener(new InputLengthListener(450));
	    input.setBackground(inputBackground);
	    input.setForeground(inputForeground);
	    add(input, gbc);
	}
	/**
	 * Used to append a String to the chat view.
	 * 
	 * @param str The String to be appended.
	 */
	public void append(String str) {
	    chatView.append(str);
	}
	/**
	 * Used to paint the background image if there is one.
	 */
	public void paint(Graphics g) {	
	    if(chatlet.background != null)
		g.drawImage(chatlet.background, 0, 0, this);
	    input.requestFocus();
	}
	/**
	 * Listens for arrow up/down to traverse the command history.
	 */
	class HistoryListener extends KeyAdapter {
	    public void keyPressed(KeyEvent evt) {
		if(evt.getKeyCode() == KeyEvent.VK_UP) {
		    if(evt.isShiftDown())
			position = history.size();
		    else
			position++;
		    input.setText(history.entryAt(position).toString());
		    input.setCaretPosition(input.getText().length());
		}
		else if(evt.getKeyCode() == KeyEvent.VK_DOWN) {
		    if(evt.isShiftDown())
			position = 0;
		    else 
			position--;
		    input.setText(history.entryAt(position).toString());
		    input.setCaretPosition(input.getText().length());
		}
	    }
	}
	/**
	 * Listens for events in the input field.
	 */
	class InputListener implements ActionListener {
	    /**
	     * Sends a message when user hits enter.
	     */
	    public void actionPerformed(ActionEvent evt) {
		input.setEnabled(false);
		input.setEditable(false);
		String message = input.getText().trim();
	        boolean error = false;
		if(message.length() > 0) {
		    message = parseOutput(message);
		    if(message != null) {		      
			try {
			    if(position > 0) 
				history.replaceEntry(position, message);
			    else
				history.addEntry(message);
			    chatlet.servlet.writeObject(chatlet.code + 
							":" + message);
			}
			catch(IOException e) {
			    chatlet.showStatus("Could not send message: " + e);
			    error = true;
			}
		    }
		}
	        if(!error) {
		   position = 0;
		   input.setText("");
		}
		input.setEditable(true);
		input.setEnabled(true);
		input.requestFocus();
	    }
	}
	/**
	 * Used to clear the chat view area and history.
	 */
	public void clear() {
	    history.clear();
	    chatView.setText(chatlet.motd + "\n");
	}
	/**
	 * Used to parse an outgoing message.
	 * 
	 * @param message The message to be parsed.
	 * @return        The parsed message to be sent or
	 *                null otherwise.
	 */
	public String parseOutput(String message) {       
	    String recipient = null;
	    String parsed;
	    if(!message.startsWith("/")) {
		parsed = "/say " + message;
	    }
	    else if(message.toLowerCase().startsWith("/clear")) {
		clear();
		parsed = null;
	    }
	    else {
		parsed = message;
	    }
	    return parsed;
	}
    }
    /**
     * This panel is used in the interface to collect the 
     * login information.
     *
     * @author Lars Samuelsson 
     */
    class LoginPanel extends Panel {
	// collects the nickname
	private TextField name;
	// collects the real name
	private TextField realname;
	// collects the server
	private TextField server;
	// collects the port number
	private TextField port;
	// the font used
	private Font typeface;
	// a reference to the chatlet
	private Chatlet chatlet;
	// the color of the text in the label
	private Color labelForeground;
	// the button to click when logging in
	private Button loginButton;
	// an offset between label and input field
	private int maxLabelWidth = 0;
	// a list that holds labels and components
	private Vector labels;
	/**
	 * This creates the LoginPanel.
	 *
	 * @param chatlet A reference to the chatlet.
	 * @param typeface The typeface to use.
	 * @param labelForeground The foreground color for the label.
	 * @param inputBackground The background color for the input field.
	 * @param inputForeground The foreground color for the input field.
	 * @param defaultServer The default server to connect to.
	 * @param serverChange Enables/disables the server field.
	 * @param defaultPort The default port to connect to.
	 * @param portChange Enables/disables the port field.
	 */
	public LoginPanel(Chatlet chatlet,
			  Font typeface,
			  Color labelForeground, 
			  Color inputBackground, 
			  Color inputForeground, 
			  String defaultServer, 
			  boolean serverChange,
			  String defaultPort,
			  boolean portChange) {
	    this.chatlet = chatlet;
	    this.labelForeground = labelForeground;
	    this.typeface = typeface;
	    name = new TextField(10);
	    name.setFont(typeface);
	    name.addTextListener(new InputLengthListener(20));
	    name.setBackground(inputBackground);
	    name.setForeground(inputForeground);
	    realname = new TextField("Chatlet User", 20);
	    realname.setFont(typeface);
	    realname.addTextListener(new InputLengthListener(40));
	    realname.setBackground(inputBackground);
	    realname.setForeground(inputForeground);
	    server = new TextField(defaultServer, 30);
	    server.setFont(typeface);
	    server.addTextListener(new InputLengthListener(50));
	    server.setBackground(inputBackground);
	    server.setForeground(inputForeground);
	    server.setEnabled(serverChange);
	    port = new TextField(defaultPort, 4);
	    port.setFont(typeface);
	    port.addTextListener(new InputLengthListener(10));
	    port.setBackground(inputBackground);
	    port.setForeground(inputForeground);
	    port.setEnabled(portChange);
	    loginButton = new Button("Login");
	    loginButton.setFont(typeface);
	    loginButton.addActionListener(new LoginListener());
	    labels = new Vector();
	    add("Nickname:", name);
	    add("Real name:", realname);
	    add("Server:", server);
	    add("Port:", port);
	    add("", loginButton);
	}
	public Component add(String label, Component comp) {
	    FontMetrics fm = comp.getFontMetrics(typeface);
	    int width = fm.stringWidth(label);
	    if(width > maxLabelWidth)
		maxLabelWidth = width;
	    super.add(comp);
	    labels.addElement(label);
	    labels.addElement(comp);
	    return comp;
	}
	/**
	 * Used to force a paint on request.
	 */
	public void update(Graphics g) {
	    paint(g);
	}
	/**
	 * This method is used to write the labels onto the 
	 * background giving an effect of transparency.
	 */
	public void paint(Graphics g) {	
	    if(chatlet.background != null)
		g.drawImage(chatlet.background, 0, 0, null);
	    g.setFont(typeface);
	    g.setColor(labelForeground);
	    Component comp;
	    String label;
	    int xl = 10;
	    int yl = 10;
	    int xc = maxLabelWidth + 8;
	    FontMetrics fm = chatlet.getFontMetrics(typeface);
	    int fontoff = fm.getHeight()/2;
	    int i = 0;
	    while(i < labels.size()) {
		label = (String)labels.elementAt(i++);
		comp = (Component)labels.elementAt(i++);
		comp.setLocation(xl + xc, yl);
		int height = comp.getSize().height;
		g.drawString(label, xl, yl + fontoff +
			     (height - fontoff)/2);
		yl += height + 4;
	    }
	}
	/**
	 * This inner class does some input checking and then 
	 * initiates the connection.
	 */
	class LoginListener implements ActionListener {
	    public void actionPerformed(ActionEvent evt) {
		if(name.getText().trim().length() < 3) {
		    chatlet.showStatus("Nickname must be at least " +
				       "three characters long");
		    return;
		}
		if(realname.getText().trim().length() < 1) {
		    chatlet.showStatus("You must provide a real name");
		    return;
		}
		if(server.getText().trim().length() < 3) {
		    chatlet.showStatus("You must specify a valid server");
		    return;
		}
		int portnum;
		try {
		    portnum = Integer.parseInt(port.getText().trim());
		}
		catch(Exception e) {
		    chatlet.showStatus("You must specify a valid port number");
		    return;
		}
	        loginButton.setEnabled(false);
		chatlet.name = name.getText().trim();
		chatlet.realname = realname.getText().trim();
		chatlet.server = server.getText().trim();
		chatlet.port = portnum;
		chatlet.showStatus("Connecting to " + chatlet.server + 
				   ":" + chatlet.port +
				   " via " + chatlet.servletURL);
		try {
		    chatlet.servlet.connect();
		}
		catch(Exception e) {
		    chatlet.showStatus("Connection failure: " + e);
		    loginButton.setEnabled(true);
		    return;
		}
		chatlet.cards.show(chatlet, "chat");
		chatlet.showStatus("Connected to " + chatlet.server + 
				   ":" + chatlet.port);
	        loginButton.setEnabled(true);
	    }
	}
    }
}
