import java.io.*;
import java.net.*;

/**
 * This class connects and maintains the connection
 * with the servlet using the HTTP protocol over 
 * port 80 and a fake keep-alive.
 *
 * Note that the fake keep-alive will use some
 * extra bandwidth, but neglectably little in 
 * comparison with the message data being sent
 * back and forth between the applet and the 
 * servlet.
 *
 * @author Lars Samuelsson
 */
public class ServletConnection implements Runnable {
    // The URL of the servlet to connect to.
    private String servletURL;
    // The input stream.
    private ObjectInputStream stream;
    // The instance that implements the ConnectionHandler interface.
    private ConnectionHandler connectionHandler;
    // A thread to constantly read data objects from the servlet.
    private Thread engine = null;
    /**
     * This creates a ServletConnection instance with the given
     * parameters. Note that a connection is not made during 
     * construction.
     *
     * @param connectionHandler The connectionHandler instance.
     * @param servletURL        The URL to connect to.
     */
    public ServletConnection(ConnectionHandler connectionHandler, 
			     String servletURL) {
	this.connectionHandler = connectionHandler;
	this.servletURL = servletURL;
    }
    /**
     * Used to connect to the servlet with the option
     * of handshaking.
     */
    public void connect() throws Exception {
	URLConnection connection = new URL(servletURL).openConnection();
	connection.setDoOutput(true);
	connection.setDoInput(true);
	connection.setUseCaches(false);
	// connection.connect();
	ObjectOutputStream client = 
	    new ObjectOutputStream(connection.getOutputStream());
	Object connectionData = connectionHandler.getConnectionData();
	if(connectionData != null) {
	    client.writeObject(connectionData);
	    client.flush();
	    client.close();
	}
	stream = new ObjectInputStream(connection.getInputStream());
	connectionHandler.setConnectionResponse(stream.readObject());
	if(connectionHandler.isConnected()) {
	    engine = new Thread(this, "Fake Keep-Alive Engine");
	    engine.start();
	}
    }
    /**
     * Used to disconnect from the servlet. 
     */
    public void disconnect() {
	connectionHandler.setConnected(false);
    }
    /**
     * In the threaded part, the fake keep-alive will 
     * constantly query the servlet for new data objects.
     * Once a data object is recieved it will be sent to 
     * the ConnectionHandler.
     */
    public void run() {
	try {
	    while(engine.isAlive() && connectionHandler.isConnected()) {
	       try {
		  connectionHandler.handleData(stream.readObject());
	       }
	       catch(ClassNotFoundException e) {
		  // disregard everything else
	       }
	    }
	}
	catch(Exception e) {
	    System.out.println("Closing connection: " + e);
	    connectionHandler.setConnected(false);
	}
        try {
	    stream.close();
	}
        catch(IOException e) {
	   // trying to close the stream nicely
	}
    }
    /**
     * Used to write a data object to the servlet.
     * 
     * @param obj The object to be sent.
     */
    public synchronized void writeObject(Object obj) throws IOException {
	URLConnection connection = new URL(servletURL).openConnection();
	connection.setDoOutput(true);
	connection.setDoInput(true);
	connection.setUseCaches(false);
	connection.setRequestProperty("Content-Type", 
				      "application/x-www-form-urlencoded");

	// connection.connect();
	ObjectOutputStream client = 
	    new ObjectOutputStream(connection.getOutputStream());
	client.writeObject(obj);
	client.flush();
	client.close();
	connection.getInputStream().close();
    }
}

