import javax.servlet.*;
import java.util.*;
import java.io.*;

/**
 * This servlet will take care of incoming users 
 * connecting through the Chatlet and open a 
 * connection to an IRC server, thereafter
 * relaying information between them.
 * 
 * @author Lars Samuelsson
 */
public class ChatServlet extends GenericServlet {
    // the delay between sent packets in the fake keep alive
    private int keepAliveDelay;
    // tables of users and their connections
    private Hashtable users;
    // a table containing banned hosts and connection attempts
    private Hashtable bans;
    // a mask for the debuglevel 
    private int debug = 0;
    // the deamon for ident requests
    private IdentDaemon ident = null;
    /**
     * Used to retrieve information from the servlet.
     *
     * @return A String containing connection the number
     *         of currently connected users.
     */
    public String getServletInfo() {
	return users.size() + " users are currently connected";
    }
    /**
     * Used to retrieve version information.
     * 
     * @return A String containing version info.
     */
    public String getVersion() {
       return "Chatlet v1.0 - http://www.alldeles.nu - " + 
	 "written by Lars Samuelsson (lars@delicate.se)";
    }
    /**
     * Invoked when the servlet is being shut down by 
     * the servlet engine or web server.
     */
    public void destroy() {
       broadcast(">>> Connection closed: Servlet shutdown");
       if(ident != null)
	 ident.close();
       Enumeration all = users.keys();
       while(all.hasMoreElements()) 
	 closeConnection((UserConnection)users.get(all.nextElement()));
       users.clear();
       System.out.println("ChatServlet - Shutdown complete");
    }
    /**
     * Invoked when the servlet is created. Some parameters
     * are read here and defaults are used if they are non-existant.
     */
    public void init() {
        System.out.println(getVersion()); 
	users = new Hashtable();	
        if(!"false".equals(getInitParameter("ident"))) {
	   try {
	      ident = new IdentDaemon(users);
	   }
	   catch(IOException e) {
	      System.err.println("Initialization error: " + e);
	   }
	}
	keepAliveDelay = 60000;
	if(getInitParameter("keepAliveDelay") != null)
	    keepAliveDelay = 
		Integer.parseInt(getInitParameter("keepAliveDelay").trim());
        System.out.println("- keepAliveDelay: " + keepAliveDelay);
	if(getInitParameter("debugLevel") != null) 
	    debug = 
		Integer.parseInt(getInitParameter("debugLevel").trim());
        System.out.println("- debugLevel: " + debug);
	bans = new Hashtable();
	if(getInitParameter("bans") != null) {
	    System.out.println("- bans: " + getInitParameter("bans"));
	    StringTokenizer st = 
		new StringTokenizer(getInitParameter("bans"), ",");
	    while(st.hasMoreTokens()) {
		bans.put(st.nextToken().trim(), new Vector());
	    }
	}
    }
    /**
     * Takes care of incoming messages and commands from users.
     *
     * @param request  The request object.
     * @param response The response object.
     */
    public void service(ServletRequest request, ServletResponse response)
	throws ServletException, IOException {	
	UserConnection activeConnection = null;
	String ip, host = "host@unknown";
	String data = null;
	boolean error = true;
	try {
	    ObjectInputStream ois = 
		new ObjectInputStream(request.getInputStream());
	    host = request.getRemoteHost();
	    ip = request.getRemoteAddr();
	    if(!isHostBanned(host) && 
	       (data = (String)ois.readObject()) != null) {
		if((debug & 2) == 1)
		    System.out.println(host + ">> " + data);
		int index = data.indexOf(':');
	        // command input
		if(index > -1) {
		   int code = Integer.parseInt(data.substring(0, index));
		   String input = data.substring(index + 1).trim();
		   findConnection(code, ip).execute(input);
		}
	        // login attempt
		else {
		   // response.setContentLength(10000000);
		   response.setContentType("application/octet-stream");
		   ObjectOutputStream oos = 
		     new ObjectOutputStream(response.getOutputStream());
		   activeConnection = connectUser(data, oos, ip, host);
		   activeConnection.keepAlive(keepAliveDelay);
		}
	    }
	    error = false;
	}
	catch(IOException e) {
	    System.err.println("Connection error (" + host + "): " + e);
	}
	catch(ClassNotFoundException e) {
	    System.err.println("Datatype error (" + host + "): " + e);
	}
	catch(IndexOutOfBoundsException e) {
	    System.err.println("Command error (" + host + "): " + e);
	}
	catch(NumberFormatException e) {
	    System.err.println("User error (" + host + "): " + e);
	}
	catch(Exception e) {
	    System.err.println("Internal error (" + host + "): ");
	    e.printStackTrace();
	    System.err.println("Data (user): " + data);
	}
	finally {
	    if(error) 
	        closeConnection(activeConnection);
	}
    }
    /**
     * Used to find a connection associated with a user. The 
     * method uses only the parts of information about a user
     * that makes him/her unique. 
     * 
     * @param code The unique code for this user.
     * @param host The host the user is connecting from.
     */
    public UserConnection findConnection(int code, String ip) 
	throws IOException {
	User uniqueUser = new User(code, ip);
	UserConnection activeConnection = 
	    (UserConnection)users.get(uniqueUser);
	if(activeConnection == null)
	    throw new IOException("User is not connected");
	return activeConnection;
    }
    /**
     * Used in the login sequence to create a user object.
     *
     * @param input      The input containing login information.
     * @param userStream The stream connected to this user.
     * @param ip         The ip the user is connecting from.
     * @param host       The host the user is connecting from.
     * @return The connection object for this user.
     */
    public UserConnection connectUser(String input, 
				      ObjectOutputStream userStream, 
				      String ip,
				      String host) 
     throws IOException, 
     IndexOutOfBoundsException, 
     NumberFormatException {
	String[] parts;
	String pattern = "* * * *";
	if((parts = UserConnection.split(input, pattern)) != null) {
	   String server = parts[0].trim();
	   int port = Integer.parseInt(parts[1].trim());
	   if(port < 1024)
	     throw new IOException("Port must be over 1024");
	   String nick = parts[2].trim();
	   String realname = parts[3].trim();
	   User user = new User(nick, realname, ip, host);
	   UserConnection activeConnection = 
	     new UserConnection(this, user, userStream);
	   users.put(user, activeConnection);
	   activeConnection.connect(server, port);
	   System.out.println("[" + user.getDate() + "]: " + 
			      user + " logged on to " + activeConnection);
	   return activeConnection;
	}
	else 
	  throw new IOException("Lacking login arguments");
    }
    /**
     * Used to get the current debug level mask. 
     *
     * @return The current debug level mask, where: 
     * <P>
     * 0000 (0) = no debug messages <BR>
     * 0001 (1) = incoming irc messages will be printed <BR>
     * 0010 (2) = incoming user messages will be printed <BR>
     * 0100 (4) = outgoing irc messages will be printed <BR>
     * 1000 (8) = outgoing user messages will be printed <BR>
     *
     * @return The debug level.
     */
    public int getDebug() {
	return debug;
    }
    /**
     * Used to check whether a connecting host is banned or not.
     *
     * @param host The host to check.
     * @return true if the host is banned.
     */
    public boolean isHostBanned(String host) {
	return bans.containsKey(host);
    }
    /**
     * Used to close a connection and log out a user from IRC
     * and from the servlet.
     * 
     * @param activeConnection The connection to close.
     */
    public synchronized void closeConnection(UserConnection activeConnection) {
       if(activeConnection != null && activeConnection.isConnected()) {
	  User dead = activeConnection.getUser();
	  users.remove(dead);
	  activeConnection.close();
	  System.out.println("[" + (new Date()) + "]: " + 
			     dead + " logged out successfully");
       }
    }
    /**
     * Used to broadcast a message to all connected users.
     *
     * @param str The String to be broadcasted.
     */
    public void broadcast(String str) {
	Enumeration all = users.keys();
	while(all.hasMoreElements()) {
	    try {
		((UserConnection)users.get(all.nextElement())).sendToUser(str);
	    }
	    catch(Exception e) {
		// skip the troublesome user
	    }
	}
    }
}
