import java.io.*;
import java.net.*;
import java.util.*;

/**
 * This class is used to answer ident requests from the
 * irc server according to the Identification Protocol. 
 * It will give each user a semi-unique ident, consisting
 * of a hexadecimal representation of the IP address the
 * user is connecting from.
 * 
 * @author Lars Samuelsson
 */
public class IdentDaemon implements Runnable {
   // contains connected users
   private Hashtable users;
   // a flag indicating servlet status
   private boolean alive;
   // the thread engine
   private Thread engine;
   // the server socket this service will accept connection on
   private ServerSocket service;
   /**
    * Creates and starts the ident service on port 113.
    * 
    * @param users The table of active users and connections.
    */
   public IdentDaemon(Hashtable users) throws IOException {
      this.users = users;
      service = new ServerSocket(113);
      System.out.println("- identd accepting connections on port 113");
      alive = true;
      engine = new Thread(this, "Ident Service Engine");
      engine.start();
   }
   /**
    * Accepts incoming requests and creates handlers for them
    */
   public void run() {
      while(alive) {
	 try {
	    new IdentHandler(service.accept());
	 }
	 catch(IOException e) {
	    // bad connection
	 }
      }
   }
   // Takes care of ident requests
   class IdentHandler implements Runnable {
      Socket s;
      Thread engine;
      public IdentHandler(Socket s) {
	 this.s = s;
	 engine = new Thread(this, "Ident Handler Engine");
	 engine.start();
      }
      public void run() {
	 // make sure no more than 16 characters are read
	 char[] request = new char[16];
	 String host = "host@unknown";
	 try {
	    // if reception is lagged 10 seconds, cut the connection
	    s.setSoTimeout(10000);
	    host = s.getInetAddress().getHostName();
	    InputStreamReader inreq = 
	      new InputStreamReader(s.getInputStream());
	    inreq.read(request, 0, request.length);
	    PrintWriter pw = new PrintWriter(s.getOutputStream());
	    pw.println(lookup(new String(request).trim()));
	    pw.flush();
	    s.close();
	 }
	 catch(SocketException e) {
	    // bad socket
	 }
	 catch(InterruptedIOException e) {
	    // read timed out
	 }
	 catch(IOException e) {
	    // bad connection
	 }
	 catch(Exception e) {
	    System.err.println("Identification error (" + host + "): ");
	    e.printStackTrace();
	 }
      }
      // performs the actual lookup
      private String lookup(String request) {
	 String[] parts;
	 String pattern = "*,*";
	 if((parts = UserConnection.split(request, pattern)) != null) {
	    try {
	       int local = Integer.parseInt(parts[0].trim());
	       int remote = Integer.parseInt(parts[1].trim());
	       UserConnection con;
	       Enumeration cons = users.elements();
	       while(cons.hasMoreElements()) {
		  con = (UserConnection)cons.nextElement();
		  if(con.getLocalPort() == local &&
		     con.getPort() == remote) {
		     return request + " : USERID : OTHER : " +
		       con.getUser().getIdent();
		  }
	       }
	       return request + " : ERROR : NO-USER";
	    }
	    catch(Exception e) {
	       return request + " : ERROR : INVALID-PORT";
	    }
	 }
	 else
	   return request + " : ERROR : UNKNOWN-ERROR"; 
      }
   }
   /**
    * Used to stop the service.
    */
   public void close() {
      alive = false;
      try {
	 service.close();
      }
      catch(IOException e) {
	 // we tried to be nice
      }
   }
}
