import java.util.Date;
import java.util.StringTokenizer;

/**
 * The user class holds some information about
 * a user in a channel. 
 *
 * @author Lars Samuelsson
 */
public class User {
    // nickname of the user
    private String name;
    // the real name
    private String realname;
    // the ip the user is connecting from
    private String ip;
    // host user is connecting from
    private String host;
    // user hash id 
    private int code;
    // the date the user logged in
    private Date date;
    // the ident string for this user
    private String ident;
    // the lag in seconds
    private int lag;
    /**
     * Creates a new user. This constructor uses all the 
     * information about a user.
     * 
     * @param name     The nickname of the user.
     * @param realname The real name of the user.
     * @param ip       The ip the user is connecting from.
     * @param host     The host the user is connecting from.
     */
    public User(String name, 
		String realname, 
		String ip, 
		String host) {
	this.name = name;
	this.realname = realname;
        this.ip = ip;
	this.host = host;
	this.date = new Date();
        this.lag = 0;
	this.code = super.hashCode();
        this.ident = createIdent(ip);
    }
    /**
     * Creates a new user. This constructor uses only the 
     * information used when comparing two users.
     * 
     * @param code The unique user identification code.
     * @param ip   The ip the user is connecting from.
     */
    public User(int code, String ip) {
	this.code = code;
	this.ip = ip;
    }
    /**
     * Creates an ident String from the given ip.
     * 
     * @param ip The ip address of the user (eg, 123.123.42.17).
     * @return   The ident String (eg, 7b7b2a11)
     */
    public static String createIdent(String ip) {
       StringTokenizer st = new StringTokenizer(ip, ".");
       String num, id = new String();
       while(st.hasMoreTokens()) {
	  num = Integer.toHexString(Integer.parseInt(st.nextToken()));
	  if(num.length() == 1)
	    num = "0" + num;
	  id += num;
       }
       return id;
    }
    /**
     * Used internally when searching for User objects.
     *
     * @return The hash code for this User object.
     */
    public int hashCode() {
	return code;
    }
    /**
     * Used to get the unique id of the user.
     *
     * @return The unique id of the user.
     */
    public int getCode() {
	return code;
    }
    /**
     * Used to get the nickname of the user.
     *
     * @return The nickname of the user.
     */
    public String getName() {
	return name;
    }
    /**
     * Used to set the nickname of the user.
     * 
     * @param nick The nickname of the user.
     */
    public void setName(String nick) {
       name = nick;
    }
    /**
     * Used to get the host from which the user is connecting.
     *
     * @return The host the user is connecting from.
     */
    public String getHost() {
	return host;
    }
    /**
     * Used to get the ip from which the user is connecting.
     * 
     * @return The ip the user is connecting from.
     */
    public String getIP() {
       return ip;
    }
    /**
     * Used to get the real name of the user. An empty String
     * is allowed for the real name.
     *
     * @return The real name of the user.
     */
    public String getRealName() {
	return realname;
    }
    /**
     * Used to get the date the user logged on.
     * 
     * @return The date the user logged on.
     */
    public Date getDate() {
	return date;
    }
    /**
     * Used to get the ident String for this user. The ident
     * is the ip in hexadecimal form.
     * 
     * @return The ip in hexadecimal form (8 chars).
     */
    public String getIdent() {
       return ident;
    }
    /**
     * Used to get the lag of the user (in seconds).
     * 
     * @return The number of seconds the user is lagged.
     */
    public int getLag() {  
       return lag;
    }
    /**
     * Used to set the lag of user (in seconds).
     * 
     * @param lag The number of seconds the user is lagged.
     */
    public void setLag(int lag) {
       this.lag = lag;
    }
    /**
     * Used to get a String representation of this User object.
     *
     * @return A String representation of this User object.
     */
    public String toString() {
	return code + ": " + name + "@" + host + " (" + realname + ")";
    }
    /**
     * Used to compare this User with another object.
     *
     * @param object If the passed object is another User
     *               they are considered to be the same
     *               user if they have the same code and
     *               connect from the same host.
     * @return true or false.
     */
    public boolean equals(Object object) {
	if(!(object instanceof User)) 
	    return super.equals(object);
	if(((User)object).getIP().equals(ip) &&
	   ((User)object).getCode() == code) 
	    return true;
	return false;	    
    }
}
