import java.io.*;
import java.net.*;
import java.util.*;

/**
 * This class manages a users connections. It makes it
 * possible to send messages to the user or to IRC and
 * it handles relaying between them as well.
 *
 * @author Lars Samuelsson
 */
public class UserConnection implements Runnable {
    // the User associated with the connection
    private User user;
    // the address of the IRC server
    private String server;
    // the port of the IRC server
    private int port;
    // the local port after connection
    private int local;
    // the channel the user is in
    private String channel;
    // the irc output stream
    private BufferedWriter ircOutputStream;
    // the irc input stream 
    private BufferedReader ircInputStream;
    // the user output stream
    private ObjectOutputStream userOutputStream;
    // a reference to the ChatServlet
    private ChatServlet servlet;
    // a thread for the relaying
    private Thread engine;
    // a flag indicating connection status
    private boolean connected = false;
    // a flag indicating if we have gotten a pong response from the user
    private boolean pong = false;
    // last time a ping was sent
    private long pingTime;
    // a handle on the socket to set read timeouts
    private Socket irc;
    /**
     * Out of protocol delimiter.
     */
    public static final char oop = (char) 1;
    /**
     * Creates the connection object. It will try to establish
     * a connection to the specified IRC server and start relaying
     * information from it back to the user.
     *
     * @param servlet          A reference to the ChatServlet.
     * @param user             The associated user.
     * @param userOutputStream The output stream to the user.
     * @param server           The address of the IRC server.
     * @param port             The port of the IRC server.
     */
    public UserConnection(ChatServlet servlet,
			  User user, 
			  ObjectOutputStream userOutputStream) {
	this.servlet = servlet;
	this.user = user;
	this.userOutputStream = userOutputStream;
    }
    /**
     * Used to check if this connection has been established.
     * 
     * @return true if connected
     */
    public boolean isConnected() {
        return connected;
    }
    /**
     * Used to connect to the given IRC server.
     * 
     * @param server The address of the IRC server.
     * @param port   The port of the IRC server.
     */
    public void connect(String server, int port) throws IOException {
	this.server = server;
	this.port = port;
	irc = new Socket(server, port);
        local = irc.getLocalPort();
	ircOutputStream = 
	    new BufferedWriter(new OutputStreamWriter(irc.getOutputStream()));
	ircInputStream = 
	    new BufferedReader(new InputStreamReader(irc.getInputStream()));
        // do not wait longer than 5 seconds to establish the connection
        irc.setSoTimeout(5000);
        sendToUser("" + user.getCode());
	connected = true;
	engine = new Thread(this, "Relay Engine");
	engine.start();
	sendToIRC("NICK " + user.getName());
	sendToIRC("USER " + user.getIdent() + " 8 * :" + user.getRealName());
    }
    /**
     * Used to keep a connection alive (fake keep-alive).
     */
    public void keepAlive(long keepAliveDelay)
     throws IOException {
	String ping = "/ping";
	while(connected) {
	   sendToUser(ping);
	   pingTime = new Date().getTime();
	   pong = false;
	   try {
	      Thread.currentThread().sleep(keepAliveDelay);
	   }
	   catch(InterruptedException e) {
	      // aint gonna happen
	   }
	   if(!pong)
	     throw new IOException("User timed out");    
	}
    }
    /**
     * Used to send a String to the IRC server.
     * 
     * @param str The String to be sent.
     */
    public synchronized void sendToIRC(String str) throws IOException {
        if(ircOutputStream == null)
	 throw new IOException("Stream is null");
	if(str != null) {
	    ircOutputStream.write(str + "\n", 0, str.length() + 1);
	    ircOutputStream.flush();
	    if((servlet.getDebug() & 4) == 1)
		System.out.println(user.getName() + ">>" +
				   server + ": " + str);
	}
    }
    /**
     * Used to send a String to the user.
     * 
     * @param str The String to be sent.
     */
    public synchronized void sendToUser(String str) throws IOException {
        if(userOutputStream == null)
	 throw new IOException("Stream is null");
        if(str != null) {
	    userOutputStream.writeObject(str + "\n");
	    userOutputStream.flush();
	    if((servlet.getDebug() & 8) == 1)
		System.out.println(user.getHost() + "<< " + str);
	}
    }
    /**
     * Here we relay everything from the IRC server to the 
     * user after it has been parsed.
     */
    public void run() {
	String input = "Nothing";
	if(ircInputStream != null && userOutputStream != null) {
	    try {
		while(connected && 
		      (input = ircInputStream.readLine()) != null) {
		    if((servlet.getDebug() & 1) == 1)
			System.out.println(user.getName() + "<<" + 
					   server + ": " + input);
		    sendToUser(parse(input));
		}
	    }
	    catch(InterruptedIOException e) {
	        System.err.println("Connection error (" +
				   user.getHost() + "): " + e);
	    }
	    catch(IOException e) {
	        // either party disconnected
	    }
	    catch(Exception e) {
	        System.err.println("Internal error (" + 
				   user.getHost() + "): ");
	        e.printStackTrace();
	        System.err.println("Data (irc): " + input);
	    }
	    finally {
	       servlet.closeConnection(this);
	    }
	}
    }
    /**
     * Used to split a string into parts according to 
     * a given pattern;
     * 
     * @param str     The string to be split.
     * @param pattern The pattern to use when splitting, 
     *                where a * symbolises a part that 
     *                should be returned. 
     * @return        An array of Strings corresponding 
     *                to the parts of the pattern, or 
     *                null if the pattern did not match.
     */
    public static String[] split(String str, String pattern) {
       boolean first = pattern.startsWith("*");
       Vector parts = new Vector();
       String delim = new String();
       String part;
       try {
	  for(int i = 0; i < pattern.length(); i++) {
	     if(pattern.charAt(i) != '*')
	       delim += pattern.charAt(i);
	     else if(delim.length() > 0) {
		if(!first) {
		   if(!str.startsWith(delim))
		     return null;
		   str = str.substring(str.indexOf(delim) + delim.length());
		   first = true;
		}
		else {
		   part = str.substring(0, str.indexOf(delim));
		   parts.addElement(part);
		   str = str.substring(str.indexOf(delim) + delim.length());
		}
		delim = new String();
	     }
	  }
	  if(delim.length() > 0) {
	     part = str.substring(0, str.indexOf(delim));
	     parts.addElement(part);
	  }
	  else {
	     parts.addElement(str);
	  }
       }
       catch(Exception e) {
	  return null;
       }
       String[] split = new String[parts.size()];
       for(int i = 0; i < parts.size(); i++) 
	  split[i] = (String)parts.elementAt(i);
       return split;
    }
    /**
     * Used to parse a String from the IRC server.
     *
     * @param str The String to be parsed.
     * @return    A parsed String or null if nothing
     *            should be sent back to the user.
     */
    public String parse(String str) throws IOException {
       String[] parts;
       String pattern;

       // command with pre colon args 
       pattern = ":*!* * * :*";
       if((parts = split(str, pattern)) != null) {
	  String nick = parts[0];
	  String ident = parts[1];
	  String cmd = parts[2];
	  String args = parts[3];
	  String s = parts[4];
	  // messages
	  if(cmd.equals("PRIVMSG")) {
	      // private to user
	      if(args.equals(user.getName())) {
		  // private user action
		  pattern = oop + "ACTION *" + oop;
		  if((parts = split(s, pattern)) != null)
		      return "** " + nick + " " + parts[0];	       
		  // ctcp version request
		  else if(s.startsWith(oop + "VERSION" + oop)) {
		      sendToIRC("NOTICE " + nick + " :" + oop +
				"VERSION " + servlet.getVersion() + oop);
		      return null;
		  }
		  // ctcp ping request
		  else if(s.startsWith(oop + "PING")) {
		     pattern = oop + "PING *" + oop;
		     if((parts = split(s, pattern)) != null) {
			try {
			   long sec = Long.parseLong(parts[0]); 
			   long tot = sec - user.getLag();
			   sendToIRC("NOTICE " + nick + " :" + oop +
				     "PING " + tot + oop);
			   return null;
			}
			catch(Exception e) {
			   // bad ping request
			}
		     }
		  }
		  // ctcp finger request
		  else if(s.startsWith(oop + "FINGER" + oop)) {
		      sendToIRC("NOTICE " + nick + " :" + oop +
				"FINGER " + user.getName() + "@" +
				user.getHost() + oop);
		      return null;
		  }
		  return "*" + nick + "* " + s;
	      }
	      // public to channel
	      else {
		  // public user action
		  pattern = oop + "ACTION *" + oop;
		  if((parts = split(s, pattern)) != null)
		      return "* " + nick + " " + parts[0];	       
		  return  "<" + nick + "> " + s;
	      }
	  }
	  // notice from user
	  else if(cmd.equals("NOTICE")) {
	      return "-" + nick + "- " + s;
	  }
       }

       // command with post colon args 
       pattern = ":*!* * :*";
       if((parts = split(str, pattern)) != null) {
	  String nick = parts[0];
	  String ident = parts[1];
	  String cmd = parts[2];
	  String s = parts[3];
	  // user joined channel
	  if(cmd.equals("JOIN")) {
	     if(user.getName().equals(nick))
	       channel = s;
	     return ">>> " + nick + " (" + ident + ") has joined " + s;
	  }
	  // user quit IRC
	  else if(cmd.equals("QUIT")) {
	     return ">>> " + nick + " (" + ident + ") quits: " + s;
	  }
	  // user changed nickname
	  else if(cmd.equals("NICK")) {
	     if(user.getName().equals(nick))
	       user.setName(s.trim());
	     return ">>> " + nick + " is now known as " + s;
	  }
       }
       
       // command with no (or late) colon args by any user
       pattern = ":*!* * *";
       if((parts = split(str, pattern)) != null) {
	  String nick = parts[0];
	  String ident = parts[1];
	  String cmd = parts[2];
	  String s = parts[3];
	  // user left channel
	  if(cmd.equals("PART")) {
	     if(user.getName().equals(nick))
	       channel = null;
	     return ">>> " + nick + " (" + ident + ") has left " + s;
	  }
	  // user changed mode
	  else if(cmd.equals("MODE")) {
	     pattern = "* *";
	     if((parts = split(s, pattern)) != null) {
		return ">>> Mode change on " + parts[0] + " by " + 
		  nick + ": " + parts[1];
	     }
	  }
	  else if(cmd.equals("KICK")) {
	     pattern = "* * :*";
	     if((parts = split(s, pattern)) != null) {
		if(user.getName().equals(parts[1])) {
		   channel = null;
		   return ">>> You have been kicked by " + nick +
		     " (" + parts[2] + ")";
		}
		else {
		   return ">>> " + parts[1] + " was kicked by " +
		     nick + " (" + parts[2] + ")";
		}
	     }
	  }
       }
       
       // server ping
       pattern = "PING :*";
       if((parts = split(str, pattern)) != null) {
	  sendToIRC("PONG :" + parts[0]);
	  return null;
       }
       
       // server message (local and remote)
       pattern = ":* * * *";
       if((parts = split(str, pattern)) != null) {
	  String serv = parts[0];
	  String code = parts[1];
	  String nick = parts[2];
	  String s = parts[3];
	  // whois lookup
	  if(code.equals("311")) {
	     pattern = "* * *:*";
	     if((parts = split(s, pattern)) != null)
	       return ">>> " + parts[0] + " is " + parts[1] + 
	       "@" + parts[2] + parts[3];
	  }
	  // whois lookup
	  else if(code.equals("319")) {
	     pattern = "* :*";
	     if((parts = split(s, pattern)) != null)
	       return ">>> " + parts[0] + " on " + parts[1];
	  }
	  // whois lookup
	  else if(code.equals("301")) {
	     pattern = "* :*";
	     if((parts = split(s, pattern)) != null)
	       return ">>> " + parts[0] + " is away: " + parts[1];
	  }
	  // whois lookup
	  else if(code.equals("312")) {
	     pattern = "* *:*";
	     if((parts = split(s, pattern)) != null)
	       return ">>> " + parts[0] + " using " + parts[1] + parts[2];
	  }
	  // whois lookup
	  else if(code.equals("317")) {
	     pattern = "* * * :*";
	     if((parts = split(s, pattern)) != null) {
		String since = parts[2];
		try {
		   since = new Date(Long.parseLong(since)*1000).toString();
		}
		catch(Exception e) {
		}
		return ">>> " + parts[0] + " has been idle " + parts[1] + 
		  " seconds, signed on " + since;
	     }	
	  }
	  // whois lookup
	  else if(code.equals("318")) {
	     pattern = "* :*";
	     if((parts = split(s, pattern)) != null)
	       return ">>> " + parts[0] + " " + parts[1];
	  }
	  // who search
	  else if(code.equals("352")) {
	     pattern = "* * * * * * :* *";
	     if((parts = split(s, pattern)) != null) 
	       return ">>> " + parts[0] + " " + parts[4] + " " +
	       parts[5] + " " + parts[1] + "@" +
	       parts[2] + " (" + parts[7] + ") ";
	  }	  
	  // list of users in a channel
	  else if(s.startsWith("=")) {
	     pattern = "*:*";
	     if((parts = split(s, pattern)) != null)
	       return ">>> Users: " + parts[1];
	  }
	  // channel topic
	  else if(code.equals("332")) {
	     pattern = "*:*";
	     if((parts = split(s, pattern)) != null)
	       return ">>> Topic: " + parts[1];
	  }
	  // who set topic and when
	  else if(code.equals("333")) {
	     pattern = "* * *";
	     if((parts = split(s, pattern)) != null) {
		String since = parts[2];
		try {
		   since = new Date(Long.parseLong(since)*1000).toString();
		}
		catch(Exception e) {
		}
		return ">>> Topic set by " + parts[1] + " on " + since;
	     }
	  }
	  // welcome message (see if nick was changed at logon)
	  else if(code.equals("001")) {
	     if(!nick.equals(user.getName())) 
		user.setName(nick);
	     // set a longer timeout now that we are connected (4 min)
	     irc.setSoTimeout(240000);
	     return ">>> " + s.substring(1);
	  }
	  // motd and similar
	  else {
	     if(s.startsWith(":"))
	       s = s.substring(1);
	     return ">>> " + s;
	  }
       }
       // command without ident (self)
       pattern = ":" + user.getName() + " * * :*";
       if((parts = split(str, pattern)) != null) {
	  if("MODE".equals(parts[0])) {
	     return ">>> Mode change: " + parts[1] + " " + parts[2];
	  }
       }

       // connection notice
       pattern = "NOTICE *:*";
       if((parts = split(str, pattern)) != null) 
	  return ">>> " + parts[1];

       return "!!! " + str;
    }
    /**
     * Used to parse and execute a command from the Chatlet.
     *
     * @param input The input the user typed in the Chatlet.
     */
    public void execute(String input) throws IOException {
       String[] parts;
       String command, argument;
       // command with argument
       if((parts = split(input, "/* *")) != null) {
	  command = parts[0].toLowerCase();
	  argument = parts[1];
	  execute(command, argument);
       }
       // command without argument
       else if ((parts = split(input, "/*")) != null) {
	  command = parts[0].toLowerCase();
	  execute(command, null);
       }
       else
	 throw new IOException("Invalid user input");
    }
    /**
     * Used to execute specific commands sent by the Chatlet user.
     *
     * @param command  The command issued by the user.
     * @param argument The argument of the command, or null if none.
     */
    public void execute(String command, String argument) 
     throws IOException {
	String[] parts;
	String pattern;
	if(command.equals("msg")) {
	   pattern = "* *";
	   if((parts = split(argument, pattern)) != null) {
	      String recipient = parts[0];
	      String message = parts[1]; 
	      sendToIRC("PRIVMSG " + recipient + " :" + message);
	      sendToUser(">" + recipient + "< " + message);
	   }
	   else 
	     sendToUser(">>> /msg <nickname> <message>");
	}
	else if(command.equals("say")) {
	   if(channel == null) {
	      sendToUser(">>> You are not in a channel.");
	      return;
	   }
	   sendToIRC("PRIVMSG " + channel + " :" + argument);
	   sendToUser("<" + user.getName() + "> " + argument);
	}
        else if(command.equals("me")) {
	   if(channel == null) {
	      sendToUser(">>> You are not in a channel.");
	      return;
	   }
	   sendToIRC("PRIVMSG " + channel + " :" + oop +
		     "ACTION " + argument + oop);
	   sendToUser("* " + user.getName() + " " + argument);
	}
        else if(command.equals("whois")) {
	   sendToIRC("WHOIS " + argument);
	   sendToUser(">>> Getting information about " + argument);
	}
	else if(command.equals("pong")) {
	   long now = new Date().getTime();
	   int lag = (int)((now - pingTime) / 1000);
	   user.setLag(lag);
	   pong = true;
	}
	else if(command.equals("join")) {
	   if(channel != null) {
	      sendToIRC("PART " + channel);
	      sendToUser(">>> Leaving " + channel);
	   }
	   sendToIRC("JOIN " + argument);
	   sendToUser(">>> Joining " + argument);
	}
	else if(command.equals("quit")) {
	   if(argument == null)
	     argument = "Chatlet client disconnected";
	   sendToIRC("QUIT :" + argument);
	   sendToUser(">>> Quitting (" + argument + ")");
	   servlet.closeConnection(this);
	}
        else if(command.equals("names")) {
	   sendToIRC("NAMES " + argument);
	   sendToUser(">>> Querying " + argument + " for users"); 
	}
        else if(command.equals("raw")) {
	   sendToIRC(argument);
	   sendToUser(">>> Sending raw: " + argument);
	}
        else if(command.equals("notice")) {
	   pattern = "* *";
	   if((parts = split(argument, pattern)) != null) {
	      sendToIRC("NOTICE " + parts[0] + " :" + parts[1]);
	      sendToUser(">>> Noticing " + parts[0] + " :" + parts[1]);
	   }
	   else
	     sendToUser(">>> /notice <nick> <message>");
	}
        else if(command.equals("nick")) {
	   sendToIRC("NICK " + argument);
	   sendToUser(">>> Changing nickname to " + argument);
	}
	else if(command.equals("mode")) {
	   sendToIRC("MODE " + argument);
	   sendToUser(">>> Setting mode: " + argument);
	}
        else if(command.equals("who")) {
	   sendToIRC("WHO " + argument);
	   sendToUser(">>> Using " + argument + " to search for users");
	}
	else if(command.equals("away")) {
	    if(argument == null) {
		sendToIRC("AWAY");
		sendToUser(">>> Disabling away message");
	    }
	    else {
		sendToIRC("AWAY :" + argument);
		sendToUser(">>> Enabling away message: " + argument);
	    }
	}
	else if(command.equals("part")) {
	   if(channel == null) {
	      sendToUser(">>> You are not in a channel.");
	      return;
	   }
	   sendToIRC("PART " + channel);
	   sendToUser(">>> Leaving " + channel);
	}
	else if(command.equals("topic")) {
	   if(channel == null) {
	      sendToUser(">>> You are not in a channel.");
	      return;
	   }
	   if(argument == null) {
	       sendToUser(">>> /topic <topic>");
	       return;
	   }
	   sendToIRC("TOPIC " + channel + " :" + argument);
	   sendToUser(">>> Setting the topic to: " + argument);
	}
        else 
	  sendToUser(">>> Unknown command: " + command);
    }
    /**
     * Used to close the connections to the IRC server and
     * to the user.
     */
    public void close() {
        connected = false;
        if(ircOutputStream != null) {
	   try {
	        userOutputStream.close();
	    }
	    catch(IOException e) {
	    }
	}
        if(ircOutputStream != null) {
	   try {
	        ircOutputStream.close();
	    }
	    catch(IOException e) {
	    }
	}
	if(ircInputStream != null) {
	    try {
	        ircInputStream.close();
	    }
	    catch(IOException e) {
	    }
	}
    }
    /**
     * Used to get the IRC server the user is connected to.
     *
     * @return The IRC server the user is connected to.
     */
    public String getServer() {
	return server;
    }
    /**
     * Used to get the port of the IRC server the user is connected to.
     *
     * @return The port of the IRC server the user is connected to.
     */
    public int getPort() {
	return port;
    }
    /**
     * Used to get the local port after a connection has been 
     * established.
     * 
     * @return The local port number.
     */
    public int getLocalPort() {
       return local;
    }
    /**
     * Used to get the user associated with this connection.
     * 
     * @return The user associated with this connection.
     */
    public User getUser() {
	return user;
    }
    /**
     * Used to retrieve a String containing connection information.
     * 
     * @return A String with connection information.
     */
    public String toString() {
        return server + ":" + port;
    }
}
